/*
  h_file.c   Nov98  JS  version 0.3
  
  this is a file for handling Htmp-File (tmp_hxy.000) 
  reasons and goals for this file:
  - 2GB file breaking => use LFS (64bit) or more blocks
  - error handling (perror nach printf???) Jun99
  - open/close
  - hidden variables (encapsulating of file-handling)
  - better reading of main-file
  Ideas:
  - chksum for tmp-file, check during read?
  - vvv==1 ? measure disk performance
  - compressed tmp_hxy.XXX ? (example: 1GB => 336MB, zip=17m  unzip=3m read=1m)
                        thor  gzip -c1 1GB => 640MB, zip=10m  unzip=3m
                        thor  gzip -c9 1GB => 579MB, zip=125m unzip=3m 
                        thor  gzip -c4 1GB => 550MB, zip=14m  unzip=3m
  - see pipe.c!!! (h_file,fopen,fclose,... & fn,pipe_ropen,... kapseln
  - repeat H*v if error!?
  
  this file is still changing Nov98,Apr99,Jun99,Nov99
  history:
   v2.0.1 - access to mcplx only via Macros M_d2f(), M_f2d()
   v2.0.0 - file-split removed (see history)
   03.02.00 - bug cplx rr => mcplx rr
   25.11.99 - bug in my_fread/write if file > 1MB beseitigt
   07.01.00 - bug auf MIPS beseitigt (high-low Format!!!)
               Vorsicht bei type-Umwandlungen (high-low/low-high format)

   ToDo:
     load_hr,save_hr
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>	/* strerror */
#include "defs.h"
#include "spins.h"
#include "h_file.h"
#include "pipe.h"
#include "block.h"
#if 1
#include <unistd.h>		/* test this with configure! */
#define SLEEP(x) sleep(x)
#else
#define SLEEP(x) {}
#endif

#ifndef HRMAX
#if HRMAX>0
#error "HRMAX!=0 if not defined?" 
#endif
#endif
#if HRMAX>0	/* table for 16bit index to Hamilton-entries */
 unsigned int hr_len[B_NUM*B_NUM];
 cplx hr[B_NUM*B_NUM][HRMAX];	/* better use realloc ? */
#endif

/* --- see h_file.c, htmpBBB.XXX for splitted file (>2GB) --- */
#define Htmp "./tmp/htmp"	/* ./tmp can be a link to /tmp/$USER */

/* public variables */

/* privat variables */
ULong h_seek[B_NUM*B_NUM];		/* file-pos: */
char  h_name[B_NUM*B_NUM][40];	/* file-name */
int   h_state[B_NUM*B_NUM];		/* !==0 stream is closed */
int   h_error=0;                /* set if H could not be written */

/* my comfortable erroroutput for misterious errors */
int my_ferror(char *msg){
  perror(msg);	/* errno_msg => stderr */
  SLEEP(1);
  return 0;
}

/* ====== diese Funktionen fuer bei zipped Hamiltonian ersetzen ====== */


/* including error-handling (public???) */
size_t my_fread( char *ptr, size_t size, size_t nmemb, int block){
  size_t n_readed=0,nf,ni,tries=0; /* sgi-CC can not add void* */
  nmemb*=size;
  if( nmemb )
  while(n_readed< size /* nmemb */ ){ /* mind. 1 element */
    ni=nmemb-n_readed;if(ni>4096)ni=4096;	/* 2^n=1MB pieces */
    nf=pipe_read(ptr,(int)ni,block);
    if(nf<=0){ E3("read nf=%d ni=%d n_readed=%d",(int)nf,(int)ni,(int)n_readed);
               my_ferror("read");if(tries++<5)continue;break; }
    n_readed+=nf; ptr+=nf; tries=0;
  } /* while */
  return n_readed/size;
}

int my_fclose(int block){
  if(h_state[block]){ pipe_close(block); }  h_state[block]=0;
  return 0;
}

/* 
   write large objects in small 1e6-pieces,
    because fwrite (probably) need lot of memory
   JS Nov99
 */
size_t my_fwrite( void *ptr, size_t size, size_t nmemb, int block){
  size_t n_written=0,nf,ni,tries=0;char *p=(char *)ptr;	/* sgi-CC can not add void* */
  nmemb*=size;
  if (h_error) return 0;
  while(n_written<nmemb){
    ni=nmemb-n_written;if(ni>4096)ni=4096;	/* 1MB pieces */
    nf=pipe_write(p,(int)ni,block);
    n_written+=nf; p+=nf;
    if(nf<ni){ ERR(620);if (h_error) break;
      if(tries++<5)continue;break; }
    tries=0;
  } /* while */
  return n_written/size;
}

int my_ropen(int dn,int block){	/* dn=Dateinummer??? */
  sprintf(h_name[block],Htmp "%03u.%03u",block,dn);
  if(vvv&16)printf("\nopen__%s",h_name[block] );
  if( pipe_ropen(h_name[block],block) ){ E1("fopen_%s",h_name[block]);return 1; } 
  h_state[block]=1;
  return 0;
}
int my_wopen(int dn,int block){	/* dn=Dateinummer??? */
  sprintf(h_name[block],Htmp "%03u.%03u",block,dn);
  if(vvv&1){
    if(dn==1 && block==0) printf(" write %s",h_name[block]);
    else                  printf(" %d",dn);
  }
  if( pipe_wopen(h_name[block],block) ) { E0("open"); h_error=1; return 1; }
  h_state[block]=2;
  return 0;
}


/* ============ stream-Funktionen - streamtype-unabhaengig! ====== */
/* ============  split H-stream into 1GB pieces ============== */

static int out_vvv=0;
/* --- read from Htmp-file,  len=0: close file --- */
ULong h_read(int block, struct thxy *hxy, ULong len){
  size_t nf; ULong rlen=0;
  /* --- close file if len=0 --- */
  if(!len){ my_fclose(block); h_seek[block]=0; return 0; }
  /* --- open file --- */
  if(!h_state[block]){ h_seek[block]=0;if( my_ropen(1,block) ) return 0; }
  /* --- read --- */
  if(vvv&16)if( out_vvv<1000 )
    printf("\nread blk=%3d seek=%9lu len=%9lu pp=%11p",
                   block,h_seek[block],len,(void *)hxy);
  rlen+=nf=my_fread((char *)hxy,Hsize,len,block);
  if(rlen<1){  E0("h_read"); }
  h_seek[block]+=nf;
  if(vvv&16)if( out_vvv++<1000 ) printf(" succes ret=%9lu", rlen);
  return rlen;
}

/* ---------------------------------------------------------------- */
/* block is the blocknumber */
ULong h_write(int block, struct thxy *hxy, ULong len){
  size_t nf;
  /* --- close file if len=0 --- */
  if(!len){ my_fclose(block); h_seek[block]=0; return 0; }
  if (h_error) return 1;
  if(!h_state[block]){
    h_seek[block]=0; if( my_wopen(1,block) ) { h_error=1; return 1; }
  }
  nf=my_fwrite((void *)hxy,Hsize,len,block);
  if(nf!=len) { ERR(620); h_error=1; return 1; }
  h_seek[block]+=len;
  return 0;
}

/* ------------------------------------------------------------------ */
/* put_h: schreibe matrix-element und dessen Spalte (Zeile codiert)    Dez99 */
/* get_h: lese matrix-element und dessen Spalte (Zeile+1 if Spalte>n1) */
/* ------------------------------------------------------------------ */
/* h_nz is external counter, 3.Argument = Blockindex ? (parallel ?) */
#if HRMAX>0
#define HR_SIZE   sizeof(HR_TYPE)
#else
#define HR_SIZE   sizeof(mcplx)
#endif
#define XY_SIZE   sizeof(XY_TYPE)  


#if HRMAX>0
/* seperate error routine, only loaded to CPU-cache in error-case */
void err_fulltable(){
    int i; unsigned int j;
    /* output maximum on information, what was went wrong? */
    printf("\nerr_fulltable: --debug--\n");
    for (i=0;i<B_NUM;i++) {
      printf(" block=%d b_hnz=%ld hr_len=%d\n hr[]=",i,b_hnz[i],hr_len[i]);
      for (j=0;j<hr_len[i] && j<100;j++) printf(" %e",D(hr[i][j]));
      if (j>=100) printf(" ...");
    }
    ERR(9100);
}
#endif

/* put_h fills a write buffer, this fkt writes the buffer out! */
int flush_h(int block){  /* store in file */
  if(h_error) return 1;
  if(b_hbl[block]) /* buffer filled? */
   if( h_write(block, b_hbuf[block], b_hbl[block]) ) { h_error=1; return 1; }
  b_hbl[block]=0; b_hbp[block]=b_hbuf[block];
  return 0;
}

/* store a matrix element (xy=yindex,r=value) into the file buffer b_hbl
 * (one buffer for each thread) and write to a file if buffer is full,
 * buffer and file size is expensive, try to spare
 */
void put_h(int block,ULong xy,cplx r){	/* Dez99 ToDo: don't use struct */
  struct thxy *ph;
  int tb=block & (B_NUM-1); /* for parallel-pc not B*B bufs are necessary */
  XY_TYPE yy=(XY_TYPE)xy;
#if HRMAX>0                 /* use a index table to spare disk place */
    /* store if not already stored, get index else */
    HR_TYPE rr;unsigned int j;
    /* Note: Norm(hr-r)<1e-13 can fail for some systems, */
    /*       but hr==r is probably unecessary strong (1.234567!=1.234566) */
    for(j=0;j<hr_len[block];j++) if(Norm(hr[block][j]-r)<=1e-13) break;
    /* here we should store max(hr[][]-r)<1e-13, but this takes time */
    if(j>=hr_len[block])	/* not found */
    {
       hr[block][j]=r;hr_len[block]++;
       if(hr_len[block]>=HRMAX) err_fulltable(); /* OVL */
    }
    rr=(HR_TYPE)j;		/* store index to value */
#else    
    mcplx rr=M_d2f(r);		/* store value direct */
#endif    
    /*
     *   h_nz  - number of hamilton-elements (until now)
     *   hmax  - memory space for h
     *   hfmax - file space for h per block
     *    replace this part, use own = to spare mem (3int type etc)
     */
    if( b_hnz[block]<hfmax ) ph=b_hbp[tb]; else ph=NULL;
    if(ph){	/* store */
#ifdef PACK_HXY /* store byte-compressed */
      memcpy(ph->xy,&yy,XY_SIZE); /* yy => xy */
#ifdef STORE_XandY
      static XY_TYPE xx=0; /* its a trick, does not work for MP! */
      if(yy&XY_FLAG){ xx++; if(xx==n1)xx=0; }    /* prepare for next line x */
      memcpy(ph->xx,&xx,XY_SIZE); /* yy => xy */
#endif
      memcpy(ph->r ,&rr,HR_SIZE);
#else
      ph->r=rr;ph->xy=xy;	/* store uncompressed */
  /* direct access to &ph via cast-operators lead to bus-errors on SGI-IRIX
     therefore we use slow array access
     unsigned int *pxy; ... pxy=(unsigned int *)&tmph.xy;*pxy=xy;
  */
#endif
      /* store in file-buffer */
      b_hbp[tb]++; b_hbl[tb]++;
      if(b_hbl[tb]>=HBLen) flush_h(tb);   /* flush buffer to disk-file */
      b_snz[block]++;
    }	
    if(  b_hnz[block]+1) b_hnz[block]++;
    if(!(b_hnz[block]+2))ERR(610);
}

int get_h(int block,ULong *xy, cplx *r){
  XY_TYPE yy;  /* must have same length!? high-low-byteorder */
  int tb=block & (B_NUM-1); /* for parallel-pc not B*B bufs are necessary */
#if HRMAX>0
  HR_TYPE rr=0l;
#else
  mcplx rr;
#endif
  if(xy==NULL || r==NULL){ h_read(block,NULL,0);b_hbl[tb]=0;return 0;} /* reset */
  if( !b_hbl[tb] ){ 
   b_hbp[tb]=b_hbuf[tb];
   b_hbl[tb]=h_read(block,b_hbuf[tb],HBLen);
   if( !b_hbl[tb] ){
     E0("darf nicht sein!");
     h_read(block,NULL,0);		/* close */
     block++;if(block>B_NUM*B_NUM)block=0;	/* sequentiell preliminary */
     b_hbl[tb]=h_read(block,b_hbuf[tb],HBLen);	/* reopen (nie der fall!) */
   }
  }
#ifdef PACK_HXY
  memcpy(&yy,b_hbp[tb]->xy,XY_SIZE); /* xy => yy */
  memcpy(&rr,b_hbp[tb]->r ,HR_SIZE);
#else
  yy=b_hbp[tb]->xy; rr=b_hbp[tb]->r;
#endif
  b_hbp[tb]++; b_hbl[tb]--;
#if HRMAX>0
  (*xy)=yy;(*r)=hr[block][rr];
#else
  (*xy)=yy;(*r)=M_f2d(rr);
#endif      
#if 0
  E3("debug-geth xy=%6lu r=%e i1=%lu",(ULong)(*xy),D(*r),(ULong)i1);
#endif  
  return 0;}

void ini_thxy(ULong n1){
  ULong i1,i2,i;int xy,j;struct thxy t1;
  for(i=0;i<B_NUM*B_NUM;i++){
#if HRMAX>0
    hr_len[i]=0;
#endif
    h_seek[i]=0;		/* file-pos: */
    h_name[i][0]=0;		/* file-name */
    h_state[i]=0;		/* file closed */
  }
  for(i1=0,i=2*n1;i!=0;i1++,i>>=8);
  for(i2=0,xy=1;xy!=0;i2++,xy<<=8);
  if(vvv)printf(" bytes(thxy.xy) %ld bytes(n1) %ld",(long)i2,(long)i1);
  if(  i1> i2)F0("xy to small, change struct thxy in spins.h");
  if(2*i1<=i2)printf("\nWarning: change thxy in spins.h to spare memory.");
  j=sizeof(int);	/* else compiler warnings - expression is const. */
  if(j!=sizeof(t1.xy))E0("sizeof int!=thxz.xy");
}

#define HR_FILE "tmp/hr_save.dat"

void close_thxy(int vvv){
  unsigned int i,hrmax;
  FILE *f1=NULL;
  /* if( save_mode & 1 ) --- small amount of data! save every time */
  { /* save data before possible crash */
    f1=fopen(HR_FILE,"wb");if(!f1)E0("error fopen"); }
  for(hrmax=i=0;i<B_NUM*B_NUM;i++){
    if(h_state[i])my_fclose(i);	/* file closed */
#if HRMAX>0
    if( hr_len[i]>hrmax ) hrmax=hr_len[i];
    if( f1 ){
      if(i<B_NUM && vvv)
        printf(((i==0)?" save hr[] len: %d":" %d"),hr_len[i]);
      fwrite(&hr_len[i],sizeof(hr_len[i]),1,f1);
      fwrite(&hr[i][0], sizeof(hr[i][0]), hr_len[i],f1); /* hr_len==0 ??? */
      fwrite(&b_hnz[i], sizeof(b_hnz[i]), B_NUM,f1);
      fwrite(&b_snz[i], sizeof(b_snz[i]), B_NUM,f1);
      fwrite(&b_sn1[i], sizeof(b_sn1[i]), B_NUM,f1);
    }
#endif
  }
  if(f1)fclose(f1);
  if(vvv){ printf(" close_thxy: max hr_num= %5d (entries/block)\n",hrmax); }
}

void restore_thxy(int vvv){
  unsigned int i,hrmax;
  FILE *f1=NULL;
  /* load data after crash */
  f1=fopen(HR_FILE,"rb");if(!f1)E0("error fopen");
  for(hrmax=i=0;i<B_NUM*B_NUM;i++){
    if(h_state[i])my_fclose(i);	/* file closed */
#if HRMAX>0
    if(vvv)printf(" load hr[%d]=",i);
    fread(&hr_len[i],sizeof(hr_len[i]),1,f1);
    fread(&hr[i][0], sizeof(hr[i][0]), hr_len[i],f1);
    fread(&b_hnz[i], sizeof(b_hnz[i]), B_NUM,f1);
    fread(&b_snz[i], sizeof(b_snz[i]), B_NUM,f1);
    fread(&b_sn1[i], sizeof(b_sn1[i]), B_NUM,f1);
    if( hr_len[i]>hrmax ) hrmax=hr_len[i];
    if(vvv)printf("%d",hr_len[i]);
#endif
  }
  if(f1)fclose(f1);
  if(vvv){ printf(" close_thxy: max hr_num= %5d (entries/block)\n",hrmax); }
}
/* ---------------------------------------------------------- */
