/*
 * parallel ausgefuehrt???
 * pipe.c
 *
 * Funktionen um komprimiert in Files zu schreiben/lesen
 *          
 *   v0.5: - fuer fork wird noch einmal voller Speicher benoetigt
 *           der bei exec freigegeben wird, failed bei Speicher < 2*data
 *           Lsg: fork vor mallocs und fork in child auf Kommando  
 *
 *           old: parent - pipe+fork+exec repeated (much mem)
 *           new: parent - pipe+fork      once     (+command pipe)
 *                 child - pipe+fork+exec repeated (less mem) (METHOD=2)
 *                OR use vfork (not implemented in nonBSD),popen(uses fork) ???
 *   v0.6 - popen (METHOD=1, old: METHOD=0) => same problem
 *          starting gzip (which needs only some kB) from spin (which
 *           has GB allocated) needs TEMPORARLY GB's for internal fork on IRIX64
 *          perror=Resource temporarily unavailable
 *   v0.7 - METHOD=2 one early fork, malloc, fork by child via commands like
 *          "r cmd"=open_pipe_for_reading, "w cmd"=open_pipe_for_writing
 *          "d xxx"=write_following_xxx_data or read_xxx_data
 *          max. 64bytes overhead
 *   v0.8 - beschleunigt (less overhead)
 *   v0.9 - da METHOD=2 auf SGI-THOR bei >2GB >100h pg mit fehlermeldung stoppt
 *          METHOD=3 using zlib.h
 *          METHOD=4 using raw-files (save)
 *   v0.10 - METHOD 0 und 2 entfernt um einfacheren 
 *         parallelisierbaren code zu schreiben (alte version in ../utils)
 *	RAM-file?
 *   v0.11 - do not print perror if errno==0
 *   v0.12 - better error handling
 */
#if 0
#define PSC "ps -Twvf"  /* ps -V 2.0.2 ps-tree */
#else
#define PSC "ps -U jschulen -lf" /* thor */ 
#endif
 /* Method 0..2 on thor mysteric errors on big programs, 3 should work */
 /* 0=pipe-(v)fork-exec, 1=popen, 2=popen_via_child, 3=zlib 4=normal */
#ifndef METHOD
#define METHOD 3
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include "error.h"

#ifdef _AIX
#include <sys/select.h>
#include <strings.h>
#endif

#include "block.h"	/* B_NUM file buffers! */

int pvv=0;	/* DEBUG-MODE parent-child-subchild communication */

/* using setvbuf ???? */
/* also wenn das nicht Uebersichtlich ist? */ 
#if   METHOD == 1
FILE *fn[B_NUM];
#define WOPENS "gzip -fcd %s"
#define ROPENS "gzip -fc1 >%s"
#define WOPEN(name)    popen(name,"w")
#define ROPEN(name)    popen(name,"r")
#define CLOSE   pclose
#define ZERO    (FILE *)0
#define READ(buf,len,fn)   fread( buf, 1, len, fn)
#define WRITE(buf,len,fn)  fwrite( buf, 1, len, fn)

/* ZLIB_EXIST is defined in Makefile and given to gcc via -D macro=value */
#elif METHOD == 3  &&  defined(ZLIB_EXIST)
#ifdef HAVE_BZLIB
  /* using bzlib (ca. 900kB of RAM used, bz2=1/3 unzip-time=4*gz) */
#include <bzlib.h>
BZFILE *fn[B_NUM];
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name)    BZ2_bzopen(name,"wb")
#define ROPEN(name)    BZ2_bzopen(name,"rb")
#define CLOSE   BZ2_bzclose
#define ZERO    0
#define READ(buf,len,fn)   BZ2_bzread( fn, buf, len)
#define WRITE(buf,len,fn)  BZ2_bzwrite( fn, buf, len)
#else
  /* using zlib (ca. 260kB of dyn. RAM used, gz=1/2) */
#include <zlib.h>
gzFile fn[B_NUM];
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name)    gzopen(name,"wb2f")
#define ROPEN(name)    gzopen(name,"rb")
#define CLOSE   gzclose
#define ZERO    0
#define READ(buf,len,fn)   gzread( fn, buf, len)
#define WRITE(buf,len,fn)  gzwrite( fn, buf, len)
#endif /* !BZLIB */
#else			/* std-file */
#undef METHOD
#define METHOD 4	/* default */
FILE *fn[B_NUM];
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name)    fopen(name,"wb")
#define ROPEN(name)    fopen(name,"rb")
#define CLOSE   fclose
#define ZERO    (FILE *)0
#define READ(buf,len,fn)   fread( buf, 1, len, fn)
#define WRITE(buf,len,fn)  fwrite( buf, 1, len, fn)
#endif

#if METHOD==1  || METHOD==3 || METHOD==4 /* popen gzopen fopen */
/* ------------------------------- close ----------------------- */
int pipe_close(int block){
 if (fn[block]) CLOSE(fn[block]); /* only close if open */
 fn[block]=ZERO;                  /* mark as closed */
 return 0;
}
/* ------------------------------- read ------------------------ */
int pipe_ropen(char *name,int block){
    char s[100];int i; sprintf(s,WOPENS,name);
    if (fn[block]) pipe_close(block); /* forgotten close? */
    for(i=0,fn[block]=NULL;i<5 && fn[block]==NULL;i++){
      fn[block]=ROPEN(s);
      if (!fn[block]) { E2("pipe_ropen %s %d",s,i);perror(" "); sleep(60); }
    } if (!fn[block])   E0("pipe_ropen failed");
    if( i>1 ) E0("pipe_ropen => ok");
    return 0;
}
int pipe_read(char *buf,int len,int block){
    int nbytes=0; if(buf)buf[0]=0;
    if (!fn[block]) return 0;
    nbytes    = (int)READ( buf, len, fn[block]);
/*
    if (nbytes != len){ if(errno) perror("pipe_read");
      nbytes += (int)READ( buf+nbytes, len-nbytes, fn[block]);
      if (nbytes != len){ if(errno) perror("pipe_read"); }
    }
*/
    return nbytes;
}
/* --------------------------- write --------------------- */
int pipe_wopen(char *name,int block){
    char s[100];int i; sprintf(s,ROPENS,name);
    if (fn[block]) pipe_close(block); /* forgotten close? */
    for(i=0,fn[block]=NULL;i<5 && fn[block]==NULL;i++){
      fn[block]=WOPEN(s);
      if (!fn[block]) { E2("pipe_wopen %s %d",s,i);perror(" "); sleep(30); }
    } 
    if (!fn[block]) { E0("pipe_wopen => failed"); return 1; }
    if( i>1 ) E0("pipe_wopen => ok");
    return 0;
}
int pipe_write(char *s,int l,int block){
    int nbytes;
    if (!fn[block]) return 0;
    nbytes=(int)WRITE(s,l,fn[block]);
    if(nbytes != l) perror("pipe_write"); return nbytes;
}
int sub_ini(){ int i;for(i=0;i<B_NUM;i++)fn[i]=ZERO;return 0;}
int sub_end(){return 0;}
/* ================================================================= */
#endif

#if 0		/* test this */
int buf[1024*1024];
int main()
{       int i,j,k;char s[100+1];
	FILE *f1;
	setvbuf(stdout,(char *)NULL,_IONBF,0);
        sub_ini();				/* do this early!!! */
        s[100]=0;buf[0]=buf[1024*1023]=1;
#define FLEN 100 /* 100x18 */
        pipe_wopen("tmp000.gz");			/* gzip */
        for(i=1;i<FLEN;i++)
        pipe_write("Das ist ein Test.\n",18);
        system(PSC";echo main");
        pipe_close();
        printf("tmp000.gz written\n"); sleep(10);
        for(k=0;k<100000;k++){
          pipe_ropen("tmp000.gz");			/* gzip */
          for(i=1;i<FLEN;i++){
            j=pipe_read(s,18);s[j]=0;		/* timeout-zeit? */
            if(j!=18) fprintf(stderr," error pipe_read(%d)=%d\n",k,j);
            printf("%s",s);
          }
          pipe_close();
        }
        sub_end();
	return 0;
}
#endif
